cdef void reduced_hessenberg_qr(int p, double[::1, :] Q, double[:, ::1] R) nogil
cdef void hessenberg_qr(int p, double[::1, :] Q, double[:, ::1] R) nogil
cdef void hessenberg_qr_update(
    int p, double[::1, :] Q, double[:, ::1] R,
    double* kappa, double* u, double* v, double* w
) nogil
cdef void qr_update(
    int p, double[::1, :] Q, double[:, ::1] R,
    double* kappa, double* u, double* v, double* w
) nogil



from typing import Union

import numpy as np
from sklearn.utils.validation import check_is_fitted
from sklearn.base import BaseEstimator

try:
    # In Scikit-Learn 0.23 SelectorMixin is moved to sklearn.feature_selection
    from sklearn.feature_selection import SelectorMixin
except ImportError:
    from sklearn.feature_selection.base import SelectorMixin

from fanok.scores import selection_fdp, selection_power


def adaptive_significance_threshold(w: np.ndarray, q: float, offset: float = 0):
    """
    Compute the data-dependent threshold \tau from the statistics
    of the original and the knockoffs features.

    :param w: Statistics satisfying the flip-sign property.
    :param q: Desired FDR
    :param offset: Offset added to the numerator of the threshold
    expression. Knockoff+ correspond to offset=1. Defaults to 0.
    """
    w_set = np.setdiff1d(np.abs(w), 0)
    # w_set = np.union1d(np.abs(w), 0)
    w = np.broadcast_to(w, (len(w_set), len(w))).T

    numerator = np.sum(w <= -w_set, axis=0) + offset
    denominator = np.sum(w >= w_set, axis=0)
    denominator = np.clip(denominator, a_min=1, a_max=None)
    quotient = numerator / denominator

    candidates = w_set[quotient <= q]
    if candidates.size > 0:
        return np.min(candidates)
    else:
        return np.inf


def knockoffs_selection_mask(
    X: np.ndarray,
    y: np.ndarray,
    knockoffs: Union[np.ndarray, callable],
    stats: Union[np.ndarray, callable],
    q: float = 0.1,
    threshold_offset: float = 0,
):
    """
    Computes the knockoff selection mask from the data samples,
    the knockoffs, and the statistics.
    """
    if callable(knockoffs):
        X_tilde = knockoffs(X)
    else:
        X_tilde = knockoffs

    if callable(stats):
        w = stats(X, X_tilde, y)
    else:
        w = stats

    threshold = adaptive_significance_threshold(w, q, offset=threshold_offset)

    return w >= threshold


class BaseKnockoffSelector(BaseEstimator, SelectorMixin):
    """
    Base class for knockoff selectors.
    Must implement the method fit.
    """

    def __init__(self):
        pass

    def fit(self, X, y):
        """
        :param X: Data samples
        :param y: Target vector
        """
        raise NotImplementedError

    def _get_support_mask(self):
        check_is_fitted(self, "mask_")
        return self.mask_

    def score(self, X, y, ground_truth):
        """
        Computes the FDP and statistical power scores
        after fitting the selector.

        :param X: Data samples
        :param y: Target vector
        :param ground_truth: Features that are truly in the model
        """
        self.fit(X, y)
        return (
            selection_fdp(self.mask_, ground_truth),
            selection_power(self.mask_, ground_truth),
        )

    def scores(self, X, y, ground_truth, repeats=10):
        fdps, powers = [], []
        for _ in range(repeats):
            fdp, power = self.score(X, y, ground_truth)
            fdps.append(fdp)
            powers.append(power)
        return fdps, powers


class KnockoffSelector(BaseKnockoffSelector):
    """
    Main knockoff selector class, implementing the straightforward
    knockoff selector procedure.

    :param knockoffs: KnockoffGenerator object that will be fitted
    :param statistics: KnockoffStatistics object
    :param alpha: Desired FDR level
    :param offset: Offset added to the numerator of the threshold
    expression. Knockoff+ correspond to offset=1. Defaults to 0.
    :param fit_generator: Whether or not the knockoff generator
    must be fitted. If it was already fitted before, this is not
    required and it may be set to False. Defaults to True.
    """

    def __init__(
        self,
        knockoffs,
        statistics,
        alpha: float = 0.1,
        offset: float = 0,
        fit_generator: bool = True,
    ):
        self.knockoffs = knockoffs
        self.statistics = statistics
        self.alpha = alpha
        self.offset = offset
        self.fit_generator = fit_generator

        super().__init__()

    def fit(self, X, y):
        """
        :param X: Data samples
        :param y: Target vector
        """
        if self.fit_generator:
            self.knockoffs.fit(X)

        self.mask_ = knockoffs_selection_mask(
            X,
            y,
            self.knockoffs,
            self.statistics,
            q=self.alpha,
            threshold_offset=self.offset,
        )

        return self



import numpy as np

from scipy.linalg.cython_blas cimport drotg, drot, dgemv, daxpy, dscal

cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef void reduced_hessenberg_qr(int p, double[::1, :] Q, double[:, ::1] R) nogil:
    cdef:
        int inc = 1
        int size = 0
        int j = 0
        double c = 0, s = 0

    for j in range(p - 1):
        drotg(&R[j, j], &R[j + 1, j], &c, &s)
        R[j + 1, j] = 0

        size = p - j - 1
        drot(&size, &R[j, j + 1], &inc, &R[j + 1, j + 1], &inc, &c, &s)

        size = j + 2
        drot(&size, &Q[0, j], &inc, &Q[0, j + 1], &inc, &c, &s)



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef void hessenberg_qr(int p, double[::1, :] Q, double[:, ::1] R) nogil:
    cdef:
        int inc = 1
        int size = 0
        int j = 0
        double c = 0, s = 0

    for j in range(p - 1):
        drotg(&R[j, j], &R[j + 1, j], &c, &s)
        R[j + 1, j] = 0

        size = p - j - 1
        drot(&size, &R[j, j + 1], &inc, &R[j + 1, j + 1], &inc, &c, &s)
        drot(&p, &Q[0, j], &inc, &Q[0, j + 1], &inc, &c, &s)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef void hessenberg_qr_update(
    int p, double[::1, :] Q, double[:, ::1] R,
    double* kappa, double* u0, double* v0, double* w0
) nogil:
    """
    Updates the QR factorization of an upper Hessenberg matrix
    (upper triangular with potential non-null entries on the lower
    diagonal).

    :param p: Size of the matrix
    :param Q: Q factor
    :param R: R factor
    """
    cdef:
        int j = 0

        int inc = 1
        char* trans = 'T'
        double alpha = 1, beta = 0

        int size = 0
        double c = 0, s = 0

    dgemv(trans, &p, &p, &alpha, &Q[0, 0], &p, &u0[0], &inc, &beta, &w0[0], &inc)
    dscal(&p, kappa, &w0[0], &inc)

    for j in range(p - 2, -1, -1):
        drotg(&w0[j], &w0[j + 1], &c, &s)
        w0[j + 1] = 0

        size = p - j
        drot(&size, &R[j, j], &inc, &R[j + 1, j], &inc, &c, &s)
        drot(&p, &Q[0, j], &inc, &Q[0, j + 1], &inc, &c, &s)

    daxpy(&p, &w0[0], &v0[0], &inc, &R[0, 0], &inc)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef void qr_update(
    int p, double[::1, :] Q, double[:, ::1] R,
    double* kappa, double* u0, double* v0, double* w0
) nogil:
    """
    Updates the QR decomposition of a matrix.
    M = QR
    M' = M + kappa * u * v^T
    Computes the factors M' = Q'*R'

    :param p: Size of the matrix
    :param Q: Q factor
    :param R: R factor
    :param kappa: Rank-1 scalar factor
    :param u: Rank-1 vector factor
    :param v: Rank-1 vector factor
    :param w0: Working array. Overwritten at the end of the function
    """
    hessenberg_qr_update(p, Q, R, kappa, u0, v0, w0)
    hessenberg_qr(p, Q, R)


cpdef void qqr_update(
    Q: np.ndarray, R: np.ndarray,
    double kappa, double[::1] u, double[::1] v, double[::1] w = None
):
    p = Q.shape[0]
    qr_update(p, Q, R, &kappa, &u[0], &v[0], &w[0])



import numpy as np

from scipy.linalg import eigh, null_space, solve

from .base import KnockoffsGenerator
from fanok.sdp import solve_full_sdp


def fixed_knockoffs_parameters(
    X: np.ndarray, sdp_mode: str = "equi", assume_centered: bool = True
):
    """
    Computes the parameters allowing the create
    fixed knockoffs from the samples.
    """
    # TODO: not centered case
    n, p = X.shape[0], X.shape[1]
    if 2 * p > n:
        raise ValueError(
            f"The number of features cannot be larger than half the "
            f"number of data points. Found X with shape {(n, p)}"
        )

    Sigma = X.T @ X
    s = solve_full_sdp(Sigma, mode=sdp_mode)
    iS_t_s = solve(Sigma, np.diag(s))

    A = 2 * np.diag(s) - s[:, None] * iS_t_s

    eig, V = eigh(A)
    eig = np.clip(eig, a_min=0, a_max=None)  # Correcting computation errors
    C = np.diag(np.sqrt(eig)) @ V.T

    U_tilde = null_space(X.T)[:, : X.shape[1]]

    return np.eye(iS_t_s.shape[0]) - iS_t_s, U_tilde @ C


def sample_fixed_knockoffs_from_parameters(X: np.ndarray, A: np.ndarray, B: np.ndarray):
    """
    Sample fixed knockoffs from the matrix parameters A and B.
    """
    return X @ A + B


def natural_fixed_knockoffs(
    X: np.ndarray, A: np.ndarray = None, B: np.ndarray = None, sdp_mode: str = "equi"
):
    """
    Sample natural fixed knockoffs (case where n >= 2p).
    """
    if A is None or B is None:
        A, B = fixed_knockoffs_parameters(X, sdp_mode=sdp_mode)

    return sample_fixed_knockoffs_from_parameters(X, A, B)


def fixed_knockoffs_extend_samples(
    X: np.ndarray, y: np.ndarray = None, noise_estimate: float = None
):
    """
    Extends the samples X and/or y so that the feature dimension
    matches the sample size.
    """
    n, p = X.shape[0], X.shape[1]

    if X is not None:
        zeros = np.zeros((2 * p - n, p))
        X = np.vstack((X, zeros))

    if y is not None:
        if noise_estimate is None:
            raise ValueError(
                f"When 2p > n you need A noise estimation in order to extend the response vector y"
            )
        y = np.concatenate((y, np.random.normal(0, noise_estimate, size=2 * p - n)))

    return X, y


def extended_fixed_knockoffs(
    X: np.ndarray,
    y: np.ndarray = None,
    A: np.ndarray = None,
    B: np.ndarray = None,
    noise_estimate: float = None,
    sdp_mode: str = "equi",
):
    """
    Create fixed knockoffs in the regime where 2p > n >= p.
    """
    n, p = X.shape[0], X.shape[1]
    if n > 2 * p:
        return natural_fixed_knockoffs(X, A, B, sdp_mode=sdp_mode)
    elif p > n:
        raise ValueError(
            "Can't generated fixed knockoffs when the feature dimension is greater than the sample size"
        )

    # Augment X to get the right dimension
    X, y = fixed_knockoffs_extend_samples(X, y, noise_estimate=noise_estimate)
    X_tilde = natural_fixed_knockoffs(X, A, B, sdp_mode=sdp_mode)

    return X, X_tilde, y


def fixed_knockoffs(
    X: np.ndarray,
    y: np.ndarray = None,
    sdp_mode: str = "equi",
    noise_estimate: float = None,
    stack: bool = True,
    A: np.ndarray = None,
    B: np.ndarray = None,
):
    """
    Fixed X_tilde work better in the case where n >= 2p.
    They may be partially extended to the case where n >= p.
    In the high-dimension regime, use Gaussian X_tilde instead.
    """
    n, p = X.shape[0], X.shape[1]

    if p > n:
        raise ValueError(
            f"The number of features cannot be larger than the "
            f"number of data points. Found X with shape {(n, p)}"
        )

    if n >= 2 * p:
        X_tilde = natural_fixed_knockoffs(X, A, B, sdp_mode=sdp_mode)
    else:
        X, X_tilde, y = extended_fixed_knockoffs(
            X, y, A, B, noise_estimate=noise_estimate, sdp_mode=sdp_mode
        )
    if stack:
        X_tilde = np.hstack((X, X_tilde))
    if y is None:
        return X_tilde
    else:
        return X_tilde, y


def are_fixed_knockoffs_valid(X: np.ndarray, X_tilde: np.ndarray, tol: float = -1e-12):
    """
    Checks if the fixed knockoffs X_tilde are valid for the samples X.

    :param X: Data samples
    :param X_tilde: Knockoff samples
    :param tol: Tolerance threshold, defaults to -1e-12
    """
    Sigma = X.T @ X
    Sigma_tilde = X_tilde.T @ X_tilde

    difference = Sigma - X.T @ X_tilde
    diagonal = np.diagonal(difference)

    return (
        np.allclose(Sigma, Sigma_tilde)
        and np.allclose(difference - np.diag(diagonal), 0)
        and np.all(diagonal >= tol)
    )


class FixedKnockoffs(KnockoffsGenerator):
    """
    Abstraction of fixed knockoffs.
    In high-dimension (more features than samples),
    you should use Gaussian knockoffs instead.

    :param sdp_mode: Defines which method is employed to solve the SDP
    """

    def __init__(
        self, sdp_mode: str = "equi", noise_estimate: float = None, stack: bool = False
    ):
        super().__init__()
        self.sdp_mode = sdp_mode
        self.noise_estimate = noise_estimate
        self.stack = stack

        self.A_ = None
        self.B_ = None

    def fit(self, X: np.ndarray, y: np.ndarray = None):
        """
        Fits the knockoff generator against the samples.

        :param X: Data samples (rows are samples, columns are features)
        """
        self.A_, self.B_ = fixed_knockoffs_parameters(X, sdp_mode=self.sdp_mode)
        return self

    def transform(self, X: np.ndarray, y: np.ndarray = None):
        """
        Creates fixed knockoff features based on the samples X.
        Should be used after the generator is fitted.

        :param X: Data samples
        :return: Fixed knockoff
        """
        return fixed_knockoffs(
            X,
            sdp_mode=self.sdp_mode,
            y=y,
            noise_estimate=self.noise_estimate,
            stack=self.stack,
            A=self.A_,
            B=self.B_,
        )



import numpy as np

from sklearn.base import BaseEstimator
from sklearn.linear_model import Lasso, LassoLars


class LarsEstimator(BaseEstimator):
    """
    Efficiently computes z = sup{lambda | beta(lambda) != 0}.
    where beta(lambda) is the coefficient vector of a Lasso model penalized
    by the coefficient lambda.
    Advantages:
        - Finds the exact solution (rather than trying a finite number of lambdas)
        - No need to tune the search space
    Can however be slow if the matrix [X, X_tilde] is large.
    """

    def __init__(self, normalize: bool = False, lambdas=None):
        super().__init__()
        self.normalize = normalize
        self.lambdas = lambdas

    def fit(self, X, y):
        if self.lambdas is None:
            lars = LassoLars(
                alpha=1e-15, normalize=self.normalize, max_iter=4 * X.shape[1]
            )
            lars.fit(X, y)
            self.coef_ = lars.alphas_[
                np.argmax(lars.coef_path_.astype(bool), axis=1) - 1
            ]
        else:
            if self.lambdas is None:
                lambdas = np.logspace(-3, 3, 100)
            coefficients_list = np.zeros(X.shape[1])
            for lam in lambdas:
                clf = Lasso(alpha=lam)
                clf.fit(X, y)
                coefficients_list = np.maximum(
                    coefficients_list, lam * clf.coef_.astype(bool)
                )
            self.coef_ = coefficients_list
        return self



import numpy as np
from scipy.linalg import svd, eigh, qr

from fanok.factor_model._shrinkage import (
    _ledoit_wolf_shrinkage,
    _shrinkage_tr_sst,
    _stochastic_lanczos_quadrature,
)


def ledoit_wolf_shrinkage(
    X: np.ndarray, s: np.ndarray = None, mode: str = "exact", m: int = 20, n_v: int = 20
):
    """
    Computes the Ledoit-Wolf optimal shrinkage coefficient from the sample X.
    It doesn't evaluate the empirical covariance Sigma.

    :param X: Data samples
    :param s:
    :param mode: Method to compute the shrinkage.
    :param m:
    :param n_v:
    """
    n, p = X.shape

    if mode == "exact":
        if s is None:
            return _ledoit_wolf_shrinkage(X)
        else:
            tr2 = np.sum(s ** 4) / n / n
    elif mode == "random":
        tr2 = _stochastic_lanczos_quadrature(X, m, n_v)
    else:
        raise ValueError(
            f"Ledoit-Wolf shrinkage mode can be either 'exact' or 'random'. Found f{mode}."
        )

    tr, sst = _shrinkage_tr_sst(X)
    d = tr2 - tr ** 2 / p
    b = sst - n * tr2

    return np.clip(b / d / n / n, a_min=0, a_max=1), tr / p


def single_step_factor_model(X: np.ndarray, rank: int, mode: str = "ledoit"):
    """
    :param X: Data samples
    :param rank: Rank approximation
    """
    # TODO: Option to do diagonal first, low-rank after.
    n = X.shape[0]

    _, lam, Vt = svd(X, full_matrices=False)

    if mode == "ledoit":
        shrinkage, mu = ledoit_wolf_shrinkage(X, s=lam)
    elif mode == "empirical":
        shrinkage, mu = 0, 0
    else:
        raise ValueError(f"mode can only be 'empirical' or 'ledoit', found {mode}")

    lam = lam[:rank]

    # The eigenvalues are shrunk in case of Ledoit-Wolf estimation
    shrunk_lam = (1 - shrinkage) * lam * lam / n + shrinkage * mu
    V = Vt.T[:, :rank]

    # Diagonal term
    d = (
        (1 - shrinkage) / n * np.sum(X * X, axis=0)
        + shrinkage * mu
        - np.sum(V * shrunk_lam * V, axis=1)
    )
    # The coordinate ascent algorithms requires positive d
    d = np.clip(d, a_min=0, a_max=None)

    return d, V, shrunk_lam


def randomized_subspace_iteration(A_dot_v, p: int, rank: int, q: int):
    """
    For a real symmetric p*p matrix A, computes a low-rank p*r matrix Q
    whose range approximates the one of A.

    :param A_dot_v: Callable taking a matrix and returning its product with A
    :param p: Size of the matrix A
    :param rank: Rank of Q
    :param q: Number of iterations to perform
    """

    # Following algorithm 4 of the paper
    # "Finding Structure With Randomness: Probabilistic Algorithms For Constructing Approximate Matrix Decompositions"
    Omega = np.random.randn(p, rank)
    Q, R = qr(A_dot_v(Omega), mode="economic")

    for _ in range(q):
        Q, R = qr(A_dot_v(Q), mode="economic")

    return Q


def randomized_symmetric_decomposition(
    A_dot_v, p: int, rank: int, q: int, over_sample: int = 10
):
    """
    For a real symmetric matrix A, computes its eigenvalue decomposition
    with randomized algorithms.

    :param A_dot_v: Callable taking a matrix and returning its product with A
    :param p: Size of the matrix A
    :param rank: Rank of Q
    :param q: Number of iterations to perform
    :param over_sample: How many more vectors than the rank to use
    for the low-rank approximation. This is essentialy usefull for
    stability purposes when the rank is low. Defaults to 10.
    """
    Q = randomized_subspace_iteration(A_dot_v, p, rank + over_sample, q)
    B = Q.T @ A_dot_v(Q)
    eig, U = eigh(B, eigvals=(B.shape[1] - rank, B.shape[1] - 1))

    return Q @ U, eig


def randomized_factor_model(
    X: np.ndarray,
    rank: int = 5,
    over_sample: int = 10,
    num_iterations: int = 2,
    shrink: bool = True,
    shrinkage_mode: str = "random",
    q: int = None,
    m: int = 20,
    n_v: int = 20,
):
    """
    :param X: Data samples
    :param rank: Rank approximation. Defaults to 5
    :param over_sample: How many more vectors than the rank to use
    for the low-rank approximation. This is essentialy usefull for
    stability purposes when the rank is low. Defaults to 10.
    :param num_iterations: How many iterations to perform in
    the alternating minimization algorithms.
    :param shrink: Whether or not to shrink the covariance matrix
    (Ledoit-Wolf estimation). This is recommended in high dimension.
    Defaults to True.
    :param shrinkage_mode: How should the optimal shrinkage coefficient
    be computed, in case of Ledoit-Wolf estimation.
    """
    n, p = X.shape

    if shrink:
        delta, mu = ledoit_wolf_shrinkage(X, mode=shrinkage_mode, m=m, n_v=n_v)
    else:
        delta, mu = 0, 0

    if q is None:
        q = 14 if rank < 0.1 * p else 8

    diag_Sigma = np.sum(X * X, axis=0) / n  # diagonal of Sigma
    A_dot_v = lambda v: X.T @ (X @ (v / n))

    d = np.zeros(p)

    for _ in range(num_iterations):
        oracle = lambda v: (1 - delta) * A_dot_v(v) + delta * mu * v - v * d[:, None]
        U, s = randomized_symmetric_decomposition(
            oracle, X.shape[1], rank, q, over_sample=over_sample
        )
        s = np.clip(s, a_min=0, a_max=None)

        diag_U = np.sum(U * s * U, axis=1)
        d = (1 - delta) * diag_Sigma + delta * mu - diag_U
        d = np.clip(d, a_min=0, a_max=None)

    return d, U, s


class FactorModel:
    """
    Abstraction of the covariance factor model.
    If the covariance (empirical or Ledoit-Wolf) is Sigma,
    computes a diagonal plus low-rank approximation of it.

    Sigma = diag(d) + U * U^T
    """

    def fit(self, X: np.ndarray):
        raise NotImplementedError

    def transform(self):
        raise NotImplementedError


class RandomizedLowRankFactorModel(FactorModel):
    """
    Randomized factor model estimation.
    Performs an alternating minimization scheme to converge
    to local optimality.

    :param rank: Rank of the approximation
    :param over_sample: How many more vectors than the rank to use
    for the low-rank approximation. This is essentialy usefull for
    stability purposes when the rank is low. Defaults to 10.
    :param num_iterations: How many iterations to perform in
    the alternating minimization algorithms.
    :param shrink: Whether or not to shrink the covariance matrix
    (Ledoit-Wolf estimation). This is recommended in high dimension.
    Defaults to True.
    :param shrinkage_mode: How should the optimal shrinkage coefficient
    be computed, in case of Ledoit-Wolf estimation.
    """

    def __init__(
        self,
        rank: int,
        over_sample: int = 10,
        num_iterations: int = 2,
        shrink: bool = True,
        shrinkage_mode: str = "random",
    ):
        self.rank = rank
        self.shrink = shrink
        self.num_iterations = num_iterations
        self.over_sample = over_sample
        self.shrinkage_mode = shrinkage_mode

    def fit(self, X: np.ndarray):
        """
        Computes the diagonal plus low-rank approximation of the
        covariance estimated from the data samples.

        :param X: Data samples
        """
        self.d_, self.U_, self.s_ = randomized_factor_model(
            X,
            rank=self.rank,
            over_sample=self.over_sample,
            num_iterations=self.num_iterations,
            shrink=self.shrink,
            shrinkage_mode=self.shrinkage_mode,
        )

    def transform(self):
        """
        Returns the covariance approximation.
        """
        return self.d_, self.U_, self.s_



import numpy as np

from scipy.linalg import cholesky, lstsq
from sklearn.covariance import empirical_covariance, ledoit_wolf, GraphicalLassoCV

from fanok.sdp import solve_full_sdp, sdp_low_rank, sdp_hybrid
from .base import KnockoffsGenerator


def estimate_covariance(X, mode: str = "wolf", assume_centered: bool = False):
    """
    Estimates the covariance matrix from the samples X.
    The estimator is either empirical, Ledoit-Wolf or graphical Lasso.

    :param X: Data samples (rows are samples, columns are features)
    :param mode: Method to estimate the covariance, defaults to "wolf"
    :param assume_centered: Whehter or not the samples are assumed to have zero mean, defaults to True
    """
    if mode == "empirical":
        return empirical_covariance(X, assume_centered=assume_centered)
    elif mode == "wolf" or mode == "ledoit":
        return ledoit_wolf(X, assume_centered=assume_centered)[0]
    elif mode == "graphical":
        return GraphicalLassoCV(assume_centered=assume_centered).fit(X).covariance_
    else:
        raise ValueError(
            f"The parameter mode can only be 'empirical', 'wolf'/'ledoit' or 'graphical'; found {mode}"
        )


def gaussian_knockoffs_sampling_parameters(
    X: np.ndarray = None,
    mu: np.ndarray = None,
    Sigma: np.ndarray = None,
    sdp_mode: str = "sdp",
    sdp_kwargs: dict = None,
    covariance_mode: str = "wolf",
    assume_centered: bool = False,
    cov_tol: float = 1e-10,
):
    """
    Estimates the distribution N(mu_tilde, Sigma_tilde) of the knockoffs
    from the distribution of the original features.

    :param X: Data samples
    :param mu: True mean of the underlying model, optional
    :param Sigma: True covariance of the underlying model, optional
    :param sdp_mode: Method to solve the SDP
    :param sdp_kwargs: Additional keyword arguments given to the SDP solver
    :param covariance_mode: Method used to estimate the covariance matrix from the samples
    :param assume_centered: Whehter or not the samples are assumed to have zero mean, defaults to True
    """
    if (mu is None or Sigma is None) and X is None:
        raise ValueError(
            "mu/Sigma are None, and X is None. Feed X to estimate mu and Sigma"
        )
    if mu is None:
        mu = np.mean(X, axis=0)
    if Sigma is None:
        Sigma = estimate_covariance(
            X, mode=covariance_mode, assume_centered=assume_centered
        )
    if sdp_kwargs is None:
        sdp_kwargs = {}

    S = solve_full_sdp(Sigma, mode=sdp_mode, return_diag=True, **sdp_kwargs)
    mul = lstsq(Sigma, S)[0]

    mu_tilde = X - (X - mu) @ mul
    Sigma_tilde = 2 * S - S @ mul

    return mu_tilde, Sigma_tilde + cov_tol * np.eye(X.shape[1])


def sample_gaussian_knockoffs(mean: np.ndarray, L: np.ndarray):
    """
    Sample Gaussian knockoffs given the mean parameter and
    the (lower) Cholesky factor of the covariance matrix.
    """
    z = np.random.normal(size=mean.shape)  # z ~ N(0, 1)
    return mean + z @ L  # ~ N(mean, L @ L.T)


def gaussian_knockoffs(
    X,
    covariance_mode: str = "wolf",
    assume_centered: bool = False,
    sdp_mode: str = "sdp",
):
    """
    Sample Gaussian knockoffs from the original samples.
    """
    mu_tilde, Sigma_tilde = gaussian_knockoffs_sampling_parameters(
        X,
        sdp_mode=sdp_mode,
        covariance_mode=covariance_mode,
        assume_centered=assume_centered,
    )
    return sample_gaussian_knockoffs(mu_tilde, cholesky(Sigma_tilde))


class GaussianKnockoffs(KnockoffsGenerator):
    """
    Gaussian knockoffs abstraction.

    :param sdp_mode: Defines which method is employed to solve the SDP
    :param sdp_kwargs: Optional dict which is fed to the SDP solver
    :param covariance_mode: Method used to estimate the covariance matrix from the samples
    :param assume_centered: Whether or not the samples should be assumed to be centered to estimate the covariance
    """

    def __init__(
        self,
        sdp_mode: str = "sdp",
        sdp_kwargs: dict = None,
        covariance_mode: str = "wolf",
        assume_centered: bool = False,
    ):
        super().__init__()
        self.sdp_mode = sdp_mode
        self.sdp_kwargs = {} if sdp_kwargs is None else sdp_kwargs
        self.covariance_mode = covariance_mode
        self.assume_centered = assume_centered

    def fit(
        self, X: np.ndarray = None, mu: np.ndarray = None, Sigma: np.ndarray = None,
    ):
        """
        Fits the knockoff generator against the samples.
        If only X is provided, the mean and the covariance of the random variable that generated
        the data will be estimated from the samples.
        If mu and Sigma are given, they will be considered as the true parameters of the model.

        :param X: Data samples (rows are samples, columns are features)
        :param mu: True mean of the underlying model
        :param Sigma: True covariance of the underlying model
        """
        self.mu_tilde_, Sigma_tilde = gaussian_knockoffs_sampling_parameters(
            X=X,
            mu=mu,
            Sigma=Sigma,
            sdp_mode=self.sdp_mode,
            sdp_kwargs=self.sdp_kwargs,
            covariance_mode=self.covariance_mode,
            assume_centered=self.assume_centered,
        )
        self.cholesky_ = cholesky(Sigma_tilde)

        return self

    def transform(self, X: np.ndarray):
        """
        Creates knockoff features based on the samples X.
        Should be used after the generator is fitted.

        :param X: Data samples
        :return: Gaussian knockoff samples
        """
        # check_is_fitted(self, ['mu_tilde_', 'cholesky_'])
        return sample_gaussian_knockoffs(self.mu_tilde_, self.cholesky_)


class LowRankGaussianKnockoffs(KnockoffsGenerator):
    """
    Low-rank Gaussian knockoffs.
    Requires a factor-model for the covariance matrix.

    :param factor_model: Object of type FactorModel capable of
    computing the low-rank approximation of the covariance Sigma.
    :param fit_factor_model: Whether or not the factor model should
    be fitted when the generator is fitted, defaults to True
    """

    def __init__(self, factor_model, fit_factor_model: bool = True):
        super().__init__()
        self.factor_model = factor_model
        self.fit_factor_model = fit_factor_model

    def fit(
        self,
        X: np.ndarray = None,
        mu: np.ndarray = None,
        d: np.ndarray = None,
        U: np.ndarray = None,
        singular_values: np.ndarray = None,
        sdp_kwargs: dict = None,
    ):
        if self.fit_factor_model:
            self.factor_model.fit(X)

        d, U, eig = self.factor_model.transform()
        singular_values = np.sqrt(eig)

        if sdp_kwargs is None:
            sdp_kwargs = {}
        s = sdp_low_rank(d, U, singular_values=singular_values, **sdp_kwargs)

        self.s_ = s
        self.d_ = d
        (self.c_, self.Z_, self.P_,) = low_rank_gaussian_knockoffs_sampling_parameters(
            d, U, singular_values * singular_values, s
        )

    def transform(self, X: np.ndarray):
        """
        Creates knockoff features based on the samples X.
        Should be used after the generator is fitted.

        :param X: Data samples
        :return: Gaussian knockoff samples
        """
        return sample_low_rank_gaussian_knockoffs(
            X, d=self.d_, s=self.s_, c=self.c_, Z=self.Z_, P=self.P_
        )


class LowRankHybridGaussianKnockoffs(KnockoffsGenerator):
    """
    Low-rank Hybrid Gaussian knockoffs.
    Requires a factor-model for the covariance matrix.

    :param factor_model: Object of type FactorModel capable of
    computing the low-rank approximation of the covariance Sigma.
    :param fit_factor_model: Whether or not the factor model should
    be fitted when the generator is fitted, defaults to True
    """

    def __init__(self,
                 factor_model,
                 fit_factor_model: bool = True,
                 covariance_mode: str = "wolf",
                 assume_centered: bool = False):
        super().__init__()
        self.factor_model = factor_model
        self.fit_factor_model = fit_factor_model
        self.covariance_mode = covariance_mode
        self.assume_centered = assume_centered

    def fit(
        self,
        X: np.ndarray = None,
        mu: np.ndarray = None,
        Sigma: np.ndarray = None,
        d: np.ndarray = None,
        U: np.ndarray = None,
        singular_values: np.ndarray = None,
        sdp_kwargs: dict = None,
        cov_tol: float = 1e-10,
    ):
        if (mu is None or Sigma is None) and X is None:
            raise ValueError(
                "mu/Sigma are None, and X is None. Feed X to estimate mu and Sigma"
            )
        if mu is None:
            mu = np.mean(X, axis=0)
        if Sigma is None:
            Sigma = estimate_covariance(
                X, mode=self.covariance_mode,
                assume_centered=self.assume_centered
            )
        if sdp_kwargs is None:
            sdp_kwargs = {}

        if self.fit_factor_model:
            self.factor_model.fit(X)

        d, U, eig = self.factor_model.transform()
        singular_values = np.sqrt(eig)

        if sdp_kwargs is None:
            sdp_kwargs = {}
        s = sdp_hybrid(Sigma, d, U, singular_values=singular_values, **sdp_kwargs)
        S = np.diag(s)
        mul = lstsq(Sigma, S)[0]

        self.mu_tilde_ = X - (X - mu) @ mul
        Sigma_tilde = 2 * S - S @ mul + + cov_tol * np.eye(X.shape[1])

        self.cholesky_ = cholesky(Sigma_tilde)

    def transform(self, X: np.ndarray):
        """
        Creates knockoff features based on the samples X.
        Should be used after the generator is fitted.

        :param X: Data samples
        :return: Gaussian knockoff samples
        """
        return sample_gaussian_knockoffs(self.mu_tilde_, self.cholesky_)

def low_rank_gaussian_knockoffs_sampling_parameters(
    d: np.ndarray, U: np.ndarray, lam: np.ndarray, s: np.ndarray
):
    """
    Computes parameters from which low-rank Gaussian knockoffs
    can be samples from the low-rank approximation of Sigma
    and the solution to the SDP.

    :param d: Diagonal term in the approximation of Sigma
    :param U: Low-rank term in the approximation of Sigma
    :param lam: Eigenvalues associated to the low-rank term
    :param s: Solution to the SDP
    """
    N = cholesky(np.linalg.inv(np.diag(1 / lam) + (U.T / d) @ U), lower=True)
    P = (U.T / d).T @ N
    Z = (P.T * s).T

    c = 2 * s - s / d * s

    return c, Z, P


def sample_low_rank_gaussian_knockoffs(
    X: np.ndarray,
    d: np.ndarray = None,
    U: np.ndarray = None,
    lam: np.ndarray = None,
    s: np.ndarray = None,
    c: np.ndarray = None,
    Z: np.ndarray = None,
    P: np.ndarray = None,
):
    """
    Samples knockoffs from the low-rank parameters (c, Z, P).
    If they are not provided, they will be computed from (d, U, lam, s).

    :param X: Data samples
    :param d: Diagonal term in the approximation of Sigma
    :param U: Low-rank term in the approximation of Sigma.
    This parameter is not required if (c, Z, P) are given.
    :param lam: Eigenvalues associated to the low-rank term
    This parameter is not required if (c, Z, P) are given.
    :param s: Solution to the SDP
    :param c: Parameter given by sample_low_rank_gaussian_knockoffs
    :param Z: Parameter given by sample_low_rank_gaussian_knockoffs
    :param P: Parameter given by sample_low_rank_gaussian_knockoffs
    """
    if c is None or Z is None or P is None:
        c, Z, P = low_rank_gaussian_knockoffs_sampling_parameters(d, U, lam, s)

    X_tilde = sample_low_rank_gaussian(X.shape[0], c, Z)

    mu = np.mean(X, axis=0)
    mu_tilde = X - (X - mu) * s / d + (X - mu) @ P @ Z.T

    return mu_tilde + X_tilde


def sample_low_rank_gaussian(n: int, c: np.ndarray, Z: np.ndarray):
    """
    Sample n vectors from the normal distribution N(0, Omega)
    where Omega = diag(c) + Z * Z^T (c is a p-dimensional vector
    and Z a p * k matrix).
    It doesn't build the matrix Omega and is much faster than computing
    the Cholesky decomposition of Omega when p is large and k small.

    :param n: How many samples
    :param c: Diagonal term in the expression of Omega
    :param Z: Low-rank term in the expression of Omega
    """
    p, rank = Z.shape
    X = np.random.randn(n, p)
    M = np.identity(rank)

    Q = np.zeros((X.shape[0], rank))
    Y = np.zeros((X.shape[0], p))

    for j in range(p):
        t = M @ Z[j, :]
        delta = c[j] + Z[j, :] @ t
        if delta > 0:
            b = t / np.sqrt(delta)
            M -= np.outer(b, b)

            Y[:, j] = np.sqrt(delta) * X[:, j] + Q @ Z[j, :]
            Q += np.outer(X[:, j], b)
        else:
            Y[:, j] = Q @ Z[j, :]

    return Y



import numpy as np

from sklearn.base import BaseEstimator


def mat_med_w(z, w):
    w_sum = np.sum(w, axis=0)
    arg = np.argsort(z, axis=0)
    w_cum_sum = np.cumsum(w[arg])
    indexes = np.argmax(w_cum_sum >= w_sum / 2, axis=0)
    zeta_bar = np.take_along_axis(z, arg, axis=0)[indexes, :]

    # TODO: Handle special case
    # for a in zeta_bar:
    #     if a <= w_sum / 2:
    #         print(a, w_sum / 2)
    #         break

    return zeta_bar


class CentroidsEstimator(BaseEstimator):
    def __init__(self, norm: str = "l1"):
        self.norm = norm.lower()
        if self.norm != "l1" and self.norm != "l2":
            raise ValueError(f"Norm may be either L1 or L2, found {norm}")
        super().__init__()

    def fit(self, X, y):
        if self.norm == "l1":
            X_p, X_m = np.mean(X[y == 0, :], axis=0), np.mean(X[y == 1, :], axis=0)
            d = X_p - X_m
            self.coef_ = 1 / 2 * d * d
        else:
            X_p = X[y == 1]
            X_m = X[y == 0]
            n_p = X_p.shape[0]
            n_m = X_m.shape[0]
            theta_p = np.median(X[y == 1], axis=0)
            theta_m = np.median(X[y == 0], axis=0)
            w = y / n_p + (1 - y) / n_m
            theta = mat_med_w(X, w)

            obj_diff = (
                np.sum(np.abs(X_p - theta_p), axis=0) / n_p
                + np.sum(np.abs(X_m - theta_m), axis=0) / n_m
            )
            obj_eq = (
                np.sum(np.abs(X_p - theta), axis=0) / n_p
                + np.sum(np.abs(X_m - theta), axis=0) / n_m
            )

            self.coef_ = obj_eq - obj_diff

        return self



import numpy as np


def selection_bool_fdp(mask: np.ndarray, true_mask: np.ndarray, q: float = 0) -> float:
    return (mask & ~true_mask).sum() / max(mask.sum() + (0 if q == 0 else 1 / q), 1)


def selection_fdp(a: np.ndarray, b: np.ndarray, q: float = 0) -> float:
    """
    Computes the false discovery proportion (FDP) of the subset
    of selected features.

    :param a: Mask array describing the selected features
    :param b: Mask array of the ground truth
    :param q: Denominator term of the modified FDP expression, defaults to 0
    """
    return selection_bool_fdp(a.astype(bool), b.astype(bool), q=q)


def selection_bool_power(mask: np.ndarray, true_mask: np.ndarray) -> float:
    return (mask & true_mask).sum() / max(true_mask.sum(), 1)


def selection_power(a: np.ndarray, b: np.ndarray) -> float:
    """
    Computes the statistical power of the subset of selected features.

    :param a: Mask array describing the selected features
    :param b: Mask array of the ground truth
    """
    return selection_bool_power(a.astype(bool), b.astype(bool))



from .base import KnockoffsGenerator
from .fixed import fixed_knockoffs, are_fixed_knockoffs_valid, FixedKnockoffs
from .gaussian import gaussian_knockoffs, GaussianKnockoffs, LowRankGaussianKnockoffs, LowRankHybridGaussianKnockoffs



__version__ = "0.0.4"


from .generation import (
    KnockoffsGenerator,
    FixedKnockoffs,
    GaussianKnockoffs,
    LowRankGaussianKnockoffs,
    LowRankHybridGaussianKnockoffs,
)
from .factor_model import RandomizedLowRankFactorModel
from .statistics import EstimatorStatistics
from .selection import KnockoffSelector



import numpy as np
from scipy.linalg import svd

cimport cython
cimport numpy as np

from cython.parallel import prange
from libc.stdlib cimport rand

from scipy.linalg.cython_blas cimport ddot, dgemv, dscal, dnrm2
from scipy.linalg.cython_lapack cimport dbdsdc

from fanok.utils._dtypes import NP_DOUBLE_D_TYPE


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef void golub_kahan(
        int n, int p, double[:, ::1] X, double[::1] u, double[::1] v,
        int max_iterations, double[::1] d_out, double[::1] b_out
) nogil:
    cdef:
        double alpha = 0
        double beta = 0

        char* trans = 'T'
        char* no_trans = 'N'
        double c = 1
        int inc = 1
        double kappa

    for k in range(max_iterations):
        kappa = -beta
        dgemv(trans, &p, &n, &c, &X[0, 0], &p, &v[0], &inc, &kappa, &u[0], &inc)  # u = X @ v - beta * u
        alpha = dnrm2(&n, &u[0], &inc)
        kappa = 1 / alpha
        dscal(&n, &kappa, &u[0], &inc)  # u /= alpha

        kappa = -alpha
        dgemv(no_trans, &p, &n, &c, &X[0, 0], &p, &u[0], &inc, &kappa, &v[0], &inc)  # v = X.T @ u - alpha * v
        beta = dnrm2(&p, &v[0], &inc)  # np.linalg.norm(v)
        kappa = 1 / beta
        dscal(&p, &kappa, &v[0], &inc)  # v /= beta

        d_out[k] = alpha
        if k < max_iterations - 1:
            b_out[k] = beta


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef __stochastic_lanczos_quadrature(
        int n, int p, double[:, ::1] X, int m, int n_v
):
    cdef:
        double trace = 0
        double sqrt_p = np.sqrt(p)
        double[::1] u = np.zeros(n, dtype=NP_DOUBLE_D_TYPE)
        double[::1] v = np.zeros(p, dtype=NP_DOUBLE_D_TYPE)
        double[::1] d = np.zeros(m, dtype=NP_DOUBLE_D_TYPE)
        double[::1] b = np.zeros(m - 1, dtype=NP_DOUBLE_D_TYPE)

        # SVD
        char* up = 'up'
        char* comp = 'I'
        double[::1] work = np.zeros(3 * m**2 + 4 * m, dtype=NP_DOUBLE_D_TYPE)
        int[::1] i_work = np.zeros(8 * m, dtype=np.intc)
        int info
        double[:, ::1] U = np.zeros((m, m), dtype=NP_DOUBLE_D_TYPE)
        double[:, ::1] Vt = np.zeros((m, m), dtype=NP_DOUBLE_D_TYPE)

    with nogil:
        for i in range(n_v):
            for j in range(p):
                if rand() % 2:
                    v[j] = 1. / sqrt_p
                else:
                    v[j] = -1. / sqrt_p

            golub_kahan(n, p, X, u, v, m, d, b)

            dbdsdc(up, comp, &m, &d[0], &b[0], &U[0, 0], &m, &Vt[0, 0], &m, NULL, NULL, &work[0], &i_work[0], &info)

            for k in range(m):
                trace += Vt[0, k]**2 * d[k]**4

    return p / n_v * trace / n / n


cpdef _stochastic_lanczos_quadrature(X, m, n_v):
    # Array must be C-contiguous, finite, and the data type C double
    X = np.ascontiguousarray(X, dtype=NP_DOUBLE_D_TYPE)
    if not np.all(np.isfinite(X)):
        raise ValueError(f'X contains NaNs or Infs')

    # Dimensions checks
    if X.ndim != 2:
        raise ValueError(f'X must be a two dimensional array')
    n, p = X.shape

    if m < 2:
        raise ValueError(f'm must be >= 2, found {m}')

    return __stochastic_lanczos_quadrature(n, p, X, m, n_v)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef __shrinkage_tr_sst(int n, int p, double[:, ::1] X):
    cdef double tr = 0
    cdef double sst = 0
    cdef double temp = 0
    cdef int inc = 1
    cdef int i, j

    with nogil:
        for i in prange(n):
            temp = ddot(&p, &X[i, 0], &inc, &X[i, 0], &inc)
            tr += temp
            sst += temp**2

    return tr / n, sst


cpdef _shrinkage_tr_sst(X):
    # Array must be C-contiguous, finite, and the data type C double
    X = np.ascontiguousarray(X, dtype=NP_DOUBLE_D_TYPE)
    if not np.all(np.isfinite(X)):
        raise ValueError(f'X contains NaNs or Infs')

    # Dimensions checks
    if X.ndim != 2:
        raise ValueError(f'X must be a two dimensional array')
    n, p = X.shape

    return __shrinkage_tr_sst(n, p, X)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef __ledoit_wolf_shrinkage(
    int n, int p, double[:, ::1] X
):
    cdef double tr = 0  # trace(Sigma)
    cdef double tr_2 = 0  # trace(Sigma^2)
    cdef double self_dot = 0
    cdef double temp
    cdef int inc = 1
    cdef double b, d
    cdef int i, j

    with nogil:
        for i in prange(n):
            for j in range(i):
                tr_2 += 2 * (ddot(&p, &X[i, 0], &inc, &X[j, 0], &inc))**2
            temp = ddot(&p, &X[i, 0], &inc, &X[i, 0], &inc)
            tr += temp
            self_dot += temp**2

    tr_2 += self_dot
    tr = tr / n
    tr_2 = tr_2 / n / n

    d = tr_2 - (tr * tr) / p
    b = (self_dot - n * tr_2) / n / n

    b = min(b, d)

    return b / d, tr / p


def _ledoit_wolf_shrinkage(
        X
):
    # Array must be C-contiguous, finite, and the data type C double
    X = np.ascontiguousarray(X, dtype=NP_DOUBLE_D_TYPE)
    if not np.all(np.isfinite(X)):
        raise ValueError(f'X contains NaNs or Infs')

    # Dimensions checks
    if X.ndim != 2:
        raise ValueError(f'X must be a two dimensional array')
    n, p = X.shape

    return __ledoit_wolf_shrinkage(n, p, X)



import warnings

import numpy as np
from scipy.linalg import eigh
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import linkage, cut_tree

from fanok.sdp._full_rank import _full_rank
from fanok.sdp._low_rank import _sdp_low_rank

try:
    import cvxpy as cp
except ImportError:
    # CVXPY isn't installed
    cp = None


def cov_to_cor(Sigma: np.ndarray):
    """
    Converts a covariance matrix to a correlation matrix.

    :param Sigma: Covariance matrix
    """
    d = 1 / np.sqrt(np.diag(Sigma))
    return d[:, None] * Sigma * d


def sdp_equi(Sigma: np.ndarray):
    """
    Returns the minimum eigenvalue of 2 * Sigma.
    This is a cheap way to find a feasible solution to the SDP
    but knockoffs generated with it might lead to a low statistical
    power.

    :param Sigma: Covariance matrix. It is scaled to a correlation matrix
    and the result is scaled back at the end.
    """
    if Sigma.shape[0] != Sigma.shape[1]:
        raise ValueError("Sigma is not a square matrix")

    cor = cov_to_cor(Sigma)

    min_eigenvalue = eigh(cor, eigvals_only=True, eigvals=(0, 0))[0]

    if min_eigenvalue < 0:
        raise ValueError("Sigma is not psd")

    return min(1, 2 * min_eigenvalue) * np.diag(Sigma)


def cvx_sdp_full(Sigma: np.ndarray, solver: str = None, clip: bool = True, **kwargs):
    """
    Solves the SDP with CVXPY.

    :param Sigma: Covariance matrix
    :param solver: Which solver to use. Defaults to SCS
    :param clip: Whether or not to clip the solution (on the correlation matrix)
    into [0, 1]. It is only supposed to fix eventual numerical approximations
    of the solver. Defaults to True
    :param kwargs: Extra keyword arguments passed to the solver
    """
    if cp is None:
        raise ImportError(
            f"CVXPY is not installed; you cannot solve the SDP with it."
            f"Instead, either solve the SDP with coordinate ascent or install CVXPY."
        )
    # Default solver; SCS is installed by default with CVXPY
    # and scales pretty well. CVXOPT is another strong option
    # but must be installed separately.
    if solver is None:
        solver = cp.SCS

    p = Sigma.shape[0]
    if p != Sigma.shape[1]:
        raise ValueError("Sigma is not a square matrix")

    cor = cov_to_cor(Sigma)

    s = cp.Variable(p)
    objective = cp.Maximize(cp.sum(s))
    constraints = [cp.diag(s) << 2 * cor, s <= 1, s >= 0]
    problem = cp.Problem(objective, constraints)

    problem.solve(solver=solver, **kwargs)
    if s.value is None:
        raise RuntimeError("CVX didn't converge")
    if clip:
        s.value = np.clip(s.value, a_min=0, a_max=1)
    s.value *= np.diag(Sigma)

    return s.value


def make_asdp_clusters(Sigma: np.ndarray, blocks: int = 2):
    """
    Approximates the covariance matrix Sigma with a block
    diagonal matrix.

    :param Sigma: Covariance matrix
    :param blocks: Number of clusters
    """
    cor = cov_to_cor(Sigma)
    dissimilarity = 1 - cor

    distances = pdist(dissimilarity)
    lkg = linkage(distances, method="ward")
    labels = np.squeeze(cut_tree(lkg, blocks))

    sub_Sigmas = []
    indices = []
    for label in range(blocks):
        ind = np.where(labels == label)[0]
        indices.append(ind)
        sub_Sigmas.append(Sigma[np.ix_(ind, ind)])

    return indices, sub_Sigmas


def bisect_solution(Sigma: np.ndarray, s: np.ndarray, gamma_tol: float):
    gamma_min, gamma_max = 0, 1
    while gamma_max - gamma_min > gamma_tol:
        gamma = (gamma_max + gamma_min) / 2
        G = 2 * Sigma - gamma * np.diag(s)
        min_eigenvalue = eigh(G, eigvals_only=True, eigvals=(0, 0))[0]
        if min_eigenvalue >= 0:
            gamma_min = gamma
        else:
            gamma_max = gamma
    return gamma_min


def asdp(Sigma: np.ndarray, blocks: int = 2, gamma_tol: float = 1e-5, **kwargs):
    """
    Solves the SDP in two steps. First the covariance is approximated with
    a block-diagonal matrix. Sub-SDPs are solved in these blocks.
    Then, a one-dimensional SDP is efficiently solved with bisection in
    order to make the solution feasible.

    :param Sigma: Covariance matrix
    :param blocks: Number of clusters
    :param gamma_tol: Tolerance threshold when solving the one-dimensional SDP
    :param kwargs: Extra keyword arguments given to the SDP solver
    """
    p = Sigma.shape[0]
    if p != Sigma.shape[1]:
        raise ValueError("Sigma is not a square matrix")

    # Clustering step and solving sub SDPs
    indices, sub_Sigmas = make_asdp_clusters(Sigma, blocks)
    s = np.zeros(p)
    for i, sub_Sigma in enumerate(sub_Sigmas):
        s[indices[i]] = solve_full_sdp(sub_Sigma, mode="sdp", **kwargs)

    gamma = bisect_solution(Sigma, s, gamma_tol=gamma_tol)

    if gamma == 0:
        warnings.warn(
            "When solving the ASDP, found gamma = 0. "
            "The knockoffs won't have any power. "
            "Consider lowering the parameter gamma_tol",
        )
    s = s * gamma

    return s


def sdp_full(
    Sigma: np.ndarray,
    max_iterations: int = None,
    lam: float = None,
    mu: float = None,
    eps=1e-5,
    return_objectives: bool = False,
):
    """
    Solves the SDP with a fast coordinate ascent algorithm.
    Wrapper of the efficient Cython implementation.

    :param Sigma: Covariance matrix
    :param max_iterations: Maximum number of coordinate cycles
    :param lam: Initial barrier coefficient parameter. Most
    you don't need to change this parameter because the default
    value automatically adapts to the problem.
    :param mu: Barrier coefficient decay parameter. Should be comprised
    between 0 and 1. Lower values are ore aggressive but might not
    converge to the optimal value (machine precision is reached faster).
    Defaults to 0.8.
    """
    cor = cov_to_cor(Sigma)

    s, objectives = _full_rank(
        cor,
        max_iterations=max_iterations,
        lam=lam,
        mu=mu,
        eps=eps,
        return_objectives=True,
    )

    s = s * np.diag(Sigma)

    if return_objectives:
        return s, objectives
    return s


def sdp_low_rank(
    d: np.ndarray,
    U: np.ndarray,
    singular_values: np.ndarray = None,
    max_iterations: int = None,
    lam: float = None,
    mu: float = None,
    eps: float = 1e-5,
    return_objectives: bool = False,
):
    """
    Solves the low-rank SDP with coordinate ascent.
    The covariance Sigma is supposed to have the special structure
    Sigma = diag(d) + U * eigs * U^T
    Wrapper of the efficient Cython implementation.

    :param d: Positive diagonal term of the factor model
    :param U: Low-rank term of the factor model
    :param singular_values: Optional singular_values of the low-rank term
    :param max_iterations: Maximum number of coordinate cycles
    :param lam: Initial barrier coefficient parameter. Most
    you don't need to change this parameter because the default
    value automatically adapts to the problem.
    :param mu: Barrier coefficient decay parameter. Should be comprised
    between 0 and 1. Lower values are ore aggressive but might not
    converge to the optimal value (machine precision is reached faster).
    Defaults to 0.8.
    :param return_objectives: Whether or not the sequences of objectives
    should be returned. Defaults to False
    """
    if singular_values is not None:
        U = U * singular_values

    # Scale Sigma to a correlation matrix
    ztz = np.sum(U * U, axis=1)
    diag_Sigma = d + ztz
    inv_sqrt = 1 / np.sqrt(diag_Sigma)
    d = d / diag_Sigma
    U = inv_sqrt[:, None] * U

    s, objectives = _sdp_low_rank(
        d,
        U,
        max_iterations=max_iterations,
        lam=lam,
        mu=mu,
        eps=eps,
        return_objectives=True,
    )

    s = s * diag_Sigma

    if return_objectives:
        return s, objectives
    return s


def sdp_hybrid(
    Sigma: np.ndarray,
    d: np.ndarray,
    U: np.ndarray,
    singular_values: np.ndarray = None,
    max_iterations: int = None,
    lam: float = None,
    mu: float = None,
    eps: float = 1e-5,
    gamma_tol=1e-5,
):
    s = sdp_low_rank(
        d, U, singular_values, max_iterations, lam, mu, eps, return_objectives=False,
    )

    gamma = bisect_solution(Sigma, s, gamma_tol=gamma_tol)

    if gamma == 0:
        warnings.warn(
            "When solving the ASDP, found gamma = 0. "
            "The knockoffs won't have any power. "
            "Consider lowering the parameter gamma_tol",
        )
    s = s * gamma

    return s


def solve_full_sdp(
    Sigma: np.ndarray, mode: str = "equi", return_diag: bool = False, **kwargs
):
    """
    Solves the SDP with one of the available methods ("equi", "sdp",
    "cvx", "asdp").

    :param Sigma: Covariance matrix
    :param mode: Method to solve the SDP
    :param return_diag: Whether to return a diagonal matrix or not
    (just the solution vector). Defaults to False.
    """
    if mode == "equi":
        res = sdp_equi(Sigma, **kwargs)
    elif mode == "sdp":
        res = sdp_full(Sigma, **kwargs)
    elif mode == "cvx":
        res = cvx_sdp_full(Sigma, **kwargs)
    elif mode == "asdp":
        res = asdp(Sigma, **kwargs)
    elif mode == "ones":
        res = np.ones(Sigma.shape[0]) * 1e-16
    elif mode == "zero":
        res = np.zeros(Sigma.shape[0])
    else:
        raise ValueError(f"Mode can be either 'equi', 'sdp' or 'zero'. Found {mode}")

    if return_diag:
        return np.diag(res)
    else:
        return res



import numpy as np

from sklearn.base import BaseEstimator


class L1MultinomialNBStatistics(BaseEstimator):
    def __init__(
        self, remove_min: bool = True, normalize: bool = False,
    ):
        super().__init__()
        self.remove_min = remove_min
        self.normalize = normalize

    def fit(self, X, y):
        if self.remove_min:
            X = X - np.min(X)
        if self.normalize:
            X = (X.T / np.linalg.norm(X, axis=1)).T

        X_p = X[y == 1]
        X_m = X[y == 0]
        f_p = np.sum(X_p, axis=0)
        f_m = np.sum(X_m, axis=0)

        q1 = (f_p + f_m) * np.log(f_p + f_m)
        alpha = np.sum(f_p) / np.sum(f_p + f_m)
        q2 = f_p * np.log(f_p / alpha) + f_m * np.log(f_m / (1 - alpha))

        self.coef_ = q2 - q1
        return self


class L1GaussianNBStatistics(BaseEstimator):
    def __init__(self):
        super().__init__()

    def fit(self, X, y):
        X_p = X[y == 1]
        X_m = X[y == 0]

        n_p = X_p.shape[0]
        n_m = X_m.shape[0]

        mean_p = np.mean(X_p, axis=0)
        mean_m = np.mean(X_m, axis=0)
        std_p = np.std(X_p, axis=0)
        std_m = np.std(X_m, axis=0)

        mean = np.mean(X, axis=0)
        std = np.std(X, axis=0)

        q1 = (
            -np.sum(np.square(X_p - mean_p) / 2 / np.square(std_p), axis=0)
            - np.sum(np.square(X_m - mean_m) / 2 / np.square(std_m), axis=0)
            - n_p * np.log(std_p)
            - n_m * np.log(std_m)
        )
        q2 = -np.sum(np.square(X - mean) / 2 / np.square(std), axis=0) - (
            n_p + n_m
        ) * np.log(std)

        self.coef_ = q1 - q2
        return self


class L1BernoulliNBEstimator(BaseEstimator):
    def __init__(self, remove_min: bool = True):
        super().__init__()
        self.remove_min = remove_min

    def fit(self, X, y):
        if self.remove_min:
            X = X - np.min(X)

        X_p = X[y == 1]
        X_m = X[y == 0]
        f_p = np.sum(X_p, axis=0)
        f_m = np.sum(X_m, axis=0)
        n_p = X_p.shape[0]
        n_m = X_m.shape[0]

        theta_p = f_p / n_p
        theta_m = f_m / n_m
        theta = (f_p + f_m) / (n_p + n_m)

        q1 = (
            f_p * np.log(theta_p)
            + (n_p - f_p) * np.log(1 - theta_p)
            + f_m * np.log(f_m / n_m)
            + (n_m - f_m) * np.log(1 - theta_m)
        )

        q2 = (f_p + f_m) * np.log(theta) + (n_p + n_m - f_p - f_m) * np.log(1 - theta)

        self.coef_ = q1 - q2
        return self



import numpy as np
from scipy.linalg import cholesky, solve_triangular

cimport cython
cimport numpy as np

from libc.math cimport sqrt
from libc.math cimport abs as cabs

from scipy.linalg.cython_blas cimport dscal, ddot, dtrsv

from fanok.utils._dtypes import NP_DOUBLE_D_TYPE
from fanok.utils._cholesky cimport __cholupdate, __choldowndate


cdef double _clip(double x, double amin, double amax) nogil:
    return min(max(x, amin), amax)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef __full_rank(
    int p,
    double[:, ::1] Sigma,
    double[:, ::1] R,
    int max_iterations,
    double lam,
    double mu,
    double lam_min
):
    objectives = [0]

    cdef:
        double prev_s_sum = 0, current_s_sum = 0
        np.ndarray[double, ndim=1] s = np.zeros(p, dtype=NP_DOUBLE_D_TYPE)
        double[::1] y_tilde = np.empty(p, dtype=NP_DOUBLE_D_TYPE)
        int inc_1 = 1
        double[::1] choldate_buffer = np.zeros(p, dtype=NP_DOUBLE_D_TYPE)
        double zero = 0
        double beta, xi, c, sj, delta
        char* low = 'L'
        char* no_trans = 'N'
        char* diag = 'N'

    for n_iter in range(max_iterations):
        prev_s_sum = current_s_sum
        current_s_sum = 0
        for j in range(p):
            with nogil:
                # Build y_tilde
                for k in range(p):
                    y_tilde[k] = Sigma[j, k]
                y_tilde[j] = 0

                # Solve triangular
                dtrsv(low, no_trans, diag, &p, &R[0, 0], &p, &y_tilde[0], &inc_1)

                beta = 4 * ddot(&p, &y_tilde[0], &inc_1, &y_tilde[0], &inc_1)
                xi = 2 * Sigma[j, j] - s[j]
                c = beta * xi / (xi + beta)

                sj = _clip(2 * Sigma[j, j] - c - lam, 0, 1)
                delta = s[j] - sj

            if delta != 0:
                choldate_buffer[j] = sqrt(cabs(delta))
                if delta > 0:
                    __cholupdate(p, R, choldate_buffer)
                else:
                    if not __choldowndate(p, R, choldate_buffer):
                        return s, objectives

                # the buffer was overwritten, nullify it
                dscal(&p, &zero, &choldate_buffer[0], &inc_1)

                s[j] = sj

            current_s_sum += sj

        objectives.append(current_s_sum)

        if lam < lam_min:
            break

        lam = lam * mu

    return s, objectives


def _full_rank(
    Sigma,
    max_iterations=None,
    lam=None,
    mu=None,
    eps=1e-5,
    return_objectives=False
):
    # Arrays must be C-contiguous, finite, and the data type C double
    Sigma = np.ascontiguousarray(Sigma, dtype=NP_DOUBLE_D_TYPE)
    if not np.all(np.isfinite(Sigma)):
        raise ValueError(f'Sigma contains NaNs or Infs')

    # Dimensions checks
    if Sigma.ndim != 2:
        raise ValueError(f'Sigma must be a one or two dimensional array')
    p = Sigma.shape[0]
    if p != Sigma.shape[1]:
        raise ValueError(f'Sigma must be a square matrix')

    R = cholesky(2 * Sigma, lower=True).T  # Upper triangular + C-contiguous

    # Default lambda, mu and max_iterations and sanity checks
    if mu is None:
        mu = 0.8
    elif mu >= 1 or mu <= 0:
        raise ValueError(f"The decay parameter mu must be between 0 and 1. Found {mu}.")
    # Find the first value of lambda such that a coordinate of s will change
    if lam is None:
        W = solve_triangular(R, np.identity(p))
        lam = (1 / np.sum(W * W, axis=1)).max()
        lam = mu * lam
    elif lam <= 0:
        raise ValueError(f"The barrier parameter lam must be positive. Found {lam}.")
    if eps < 0:
        raise ValueError(f"eps cannot be negative. Found {eps}")
    if max_iterations is None:
        if eps == 0:
            raise ValueError(
                f"Can't automatically set the maximum number of iterations if eps is null"
            )
        max_iterations = np.ceil(np.log(eps / (3 * lam)) / np.log(mu)).astype(int) + 1
    lam_min = eps / 3  # Average eps. Default global one, divide by p too

    # Cython fast solver
    s, objectives = __full_rank(p, Sigma, R, max_iterations, lam, mu, lam_min)

    if return_objectives:
        return s, objectives
    else:
        return s



import numpy as np
from scipy.linalg import qr, solve_triangular

cimport cython
cimport numpy as np

from scipy.linalg.cython_blas cimport dgemv, ddot, dtrsv

from fanok.utils._dtypes import NP_DOUBLE_D_TYPE
from fanok.utils._qr cimport qr_update


cdef double _clip(double x, double amin, double amax) nogil:
    return min(max(x, amin), amax)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef sdp_rank_k(
    int p,
    int k,
    double[::1] d,
    double[:, ::1] U,
    double[::1] diag_Sigma,
    double[::1, :] Q,
    double[:, ::1] R,
    int max_iterations,
    double lam,
    double mu,
    double lam_min
):
    objectives = [0]

    cdef:
        np.ndarray[double, ndim=1] s = np.zeros(p, dtype=NP_DOUBLE_D_TYPE)
        double prev_s_sum = 0, current_s_sum = 0

        double a = 0, b = 0
        double kappa = 0
        double prev_sj = 0
        double[::1] qr_work = np.zeros(k, dtype=NP_DOUBLE_D_TYPE)
        double[::1] y = np.zeros(k, dtype=NP_DOUBLE_D_TYPE)
        double* u = NULL

        # BLAS commonly used parameters
        double zero = 0, one = 1
        int inc_1 = 1
        char* low = 'L'
        char* trans = 'T'
        char* not_unit = 'N'

    for i in range(max_iterations):
        prev_s_sum = current_s_sum
        current_s_sum = 0
        for j in range(p):
            # u = U[j, :]
            u = &U[j, 0]

            # y = Q.T @ u
            dgemv(trans, &k, &k, &one, &Q[0, 0], &k, u, &inc_1, &zero, &y[0], &inc_1)

            # y = (Q R)^{-1} u
            dtrsv(low, trans, not_unit, &k, &R[0, 0], &k, &y[0], &inc_1)

            # a = u^T (Q R)^{-1} u
            a = ddot(&k, &y[0], &inc_1, u, &inc_1)

            b = 2 * (
                (2 * d[j] - s[j]) * (d[j] - diag_Sigma[j]) - (s[j] - 2 * diag_Sigma[j]) * a
            ) / (2 * d[j] - s[j] - 2 * a)

            prev_sj = s[j]
            s[j] = _clip(2 * diag_Sigma[j] - lam + b, 0, 1)

            if 2 * d[j] - s[j] == 0:
                return s, objectives

            if s[j] != prev_sj:
                kappa = 2 * (s[j] - prev_sj) / (2 * d[j] - s[j]) / (2 * d[j] - prev_sj)
                qr_update(k, Q, R, &kappa, u, u, &qr_work[0])

            current_s_sum += s[j]

        objectives.append(current_s_sum)

        lam = lam * mu
        if lam < lam_min:
            break

    return s, objectives


def _sdp_low_rank(
    d,
    U,
    max_iterations=None,
    lam=None,
    mu=None,
    eps=1e-5,
    return_objectives=False
):
    # Arrays must be C-contiguous, finite, and the data type C double
    d = np.ascontiguousarray(d, dtype=NP_DOUBLE_D_TYPE)
    U = np.ascontiguousarray(U, dtype=NP_DOUBLE_D_TYPE)
    if not np.all(np.isfinite(d)):
        raise ValueError(f'The diagonal D contains NaNs or Infs')
    if not np.all(np.isfinite(U)):
        raise ValueError(f'The low rank U matrix contains NaNs or Infs')

    # Dimensions checks
    if d.ndim != 1:
        raise ValueError(f'The diagonal D must be a one-dimensional array')
    p = d.shape[0]
    if U.ndim == 1:
        U = np.reshape(U, (-1, 1))
    elif U.ndim != 2:
        raise ValueError(f'The low rank U matrix must be a one or two dimensional array')
    k = U.shape[1]
    if p != U.shape[0]:
        raise ValueError(f'Dimensions between D and U don\'t match')

    ztz = np.sum(U * U, axis=1)
    diag_Sigma = d + ztz
    Q, R = qr(np.eye(k) + (U.T / d) @ U)

    # Default lambda, mu and max_iterations
    if mu is None:
        mu = 0.8
    elif mu >= 1 or mu <= 0:
        raise ValueError(f"The decay parameter mu must be between 0 and 1. Found {mu}.")
    # Find the first value of lambda such that a coordinate of s will change
    if lam is None:
        diag_inv_2Sigma = 0.5 / d - 0.5 * np.sum((U @ solve_triangular(R, np.eye(k))) * (U @ Q), axis=1) / d / d
        lam = (1 / diag_inv_2Sigma).max()
        lam = mu * lam
    elif lam <= 0:
        raise ValueError(f"The barrier parameter lam must be positive. Found {lam}.")
    if eps < 0:
        raise ValueError(f"eps cannot be negative. Found {eps}")
    if max_iterations is None:
        if eps == 0:
            raise ValueError(
                f"Can't automatically set the maximum number of iterations if eps is null"
            )
        max_iterations = np.ceil(np.log(eps / (3 * lam)) / np.log(mu)).astype(int) + 1
    lam_min = eps / 3  # Average eps. Default global one, divide by p too

    # Cython fast solver
    s, objectives = sdp_rank_k(
        p, k, d, U, diag_Sigma, Q, R, max_iterations, lam, mu, lam_min
    )

    if return_objectives:
        return s, objectives
    else:
        return s



import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin


class KnockoffsGenerator(BaseEstimator, TransformerMixin):
    """
    Base class for all knockoff generators.
    They must implement the functions fit and transform.
    """

    def __init__(self):
        pass

    def fit(self, X: np.ndarray, y: np.ndarray = None):
        raise NotImplementedError

    def transform(self, X: np.ndarray, y: np.ndarray = None):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        return self.transform(*args, **kwargs)



import numpy as np

cimport cython

from scipy.linalg.cython_blas cimport drotg, drot, dtrsv, dscal, daxpy, ddot
from scipy.linalg.cython_lapack cimport dgesv

from libc.math cimport sqrt
from libc.math cimport abs as cabs


FLOAT_NUMPY_TYPE_MAP = {
    4 : np.float32,
    8 : np.float64
}
NP_DOUBLE_D_TYPE = FLOAT_NUMPY_TYPE_MAP[sizeof(double)]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef void __cholupdate(int p, double[:, ::1] R, double[::1] u) nogil:
    """
    Positive Cholesky rank-1 update.
    R must be upper-triangular and C-contiguous.
    At the end, u is overwritten.
    """
    cdef:
        int inc = 1
        int size = 0
        double c = 0, s = 0
        int j = 0

    for j in range(p):
        drotg(&R[j, j], &u[j], &c, &s)
        size = p - j - 1
        if size > 0:
            drot(&size, &R[j, j + 1], &inc, &u[j + 1], &inc, &c, &s)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef bint __choldowndate(int p, double[:, ::1] R, double[::1] u) nogil:
    """
    Negative Cholesky rank-1 update.
    Returns False if the update failed (matrix not psd), True otherwise.
    R must be upper-triangular and C-contiguous.
    At the end, u is overwritten.
    """
    cdef:
        int inc = 1
        int size = 1
        int k = 0
        double r = 0
        double c1 = 0, c2 = 0

    for k in range(p):
        r = (R[k, k] - u[k]) * (R[k, k] + u[k])
        if r <= 0:
            return False
        r = sqrt(r)

        size = p - k - 1

        if size > 0:
            c1 = R[k, k] / r
            c2 = -u[k] / r
            dscal(&size, &c1, &R[k, k + 1], &inc)
            daxpy(&size, &c2, &u[k + 1], &inc, &R[k, k + 1], &inc)

            c1 = r / R[k, k]
            c2 = -u[k] / R[k, k]
            dscal(&size, &c1, &u[k + 1], &inc)
            daxpy(&size, &c2, &R[k, k + 1], &inc, &u[k + 1], &inc)

        R[k, k] = r

    return True


def _check_R_u(R: np.ndarray, u: np.ndarray, upper: bool = True):
    if not upper:
        R = R.T
    R = np.ascontiguousarray(R, dtype=NP_DOUBLE_D_TYPE)
    u = np.ascontiguousarray(u, dtype=NP_DOUBLE_D_TYPE)

    p = R.shape[0]
    if R.shape[1] != p:
        raise ValueError(f'R must be a square matrix')
    if u.shape[0] != p:
        raise ValueError(f'Dimensions of R and u must agree')

    return p, R, u


def _cholupdate(R: np.ndarray, u: np.ndarray, upper: bool = True):
    p, R, u = _check_R_u(R, u, upper)
    __cholupdate(p, R, u)


def _choldowndate(R: np.ndarray, u: np.ndarray, upper: bool = True):
    p, R, u = _check_R_u(R, u, upper)
    return __choldowndate(p, R, u)



import numpy as np
from setuptools import find_packages, setup
from Cython.Build import cythonize
from distutils.extension import Extension


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="fanok",
    version="0.0.4",
    packages=find_packages(),
    license="MIT",
    python_requires=">=3.5.0",
    description="Knockoffs in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Armin Askari, Quentin Rebjock, Laurent El Ghaoui, and Alexandre d’Aspremont",
    author_email="aaskari@berkeley.edu, quentin.rebjock@gmail.com, elghaoui@berkeley.edu, aspremon@ens.fr",
    url="https://github.com/qrebjock/fanok",
    keywords=["knockoffs", "fdr", "statistics"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    ext_modules=cythonize(
        [
            Extension("fanok.utils._dtypes", ["fanok/utils/_dtypes.pyx"]),
            Extension("fanok.utils._cholesky", ["fanok/utils/_cholesky.pyx"]),
            Extension("fanok.utils._qr", ["fanok/utils/_qr.pyx"]),
            Extension("fanok.sdp._low_rank", ["fanok/sdp/_low_rank.pyx"]),
            Extension("fanok.sdp._full_rank", ["fanok/sdp/_full_rank.pyx"]),
            Extension(
                "fanok.factor_model._shrinkage",
                ["fanok/factor_model/_shrinkage.pyx"],
                # extra_compile_args=["-fopenmp"],
                # extra_link_args=["-fopenmp"],
            ),
        ],
        compiler_directives={"language_level": "3"},
        annotate=False,
    ),
    include_dirs=[np.get_include()],
)



import numpy as np

from fanok.utils._cholesky import _cholupdate, _choldowndate


def cholupdate(R: np.ndarray, u: np.ndarray, upper: bool = True):
    _cholupdate(R, u, upper)


def choldowndate(R: np.ndarray, u: np.ndarray, upper: bool = True):
    return _choldowndate(R, u, upper)



import numpy as np

from sklearn.linear_model import LassoCV


class KnockoffStatistics:
    """
    Abstraction of flip-sign knockoff statistics.

    Must implement the method evaluate.
    """

    def __init__(self):
        pass

    def evaluate(self, X, X_tilde, y):
        """
        :param X: Data samples
        :param X_tilde: Knockoff samples
        :param y: Target vector
        """
        pass

    def __call__(self, X, X_tilde, y):
        return self.evaluate(X, X_tilde, y)


class PlainKnockoffStatistics(KnockoffStatistics):
    """
    Flip-sign statistics directly computed from the concatenated
    feature matrix [X, \tilde{X}], without the use of an
    antisymmetric function.
    """

    def __init__(self):
        super().__init__()

    def evaluate(self, X, X_tilde, y):
        pass


def antisymmetric_knockoff_statistics(
    z: np.ndarray, z_tilde: np.ndarray = None, mode: str = "difference"
):
    """
    Applies an anti-symmetric function to pairs of original and knockoff
    statistics. This is a natural way to obtain statistics satisfying
    the flip sign property that is required to control the FDR.

    :param z: Statistics from the original features
    :param z_tilde: Statistics from the knockoff features.
    If None, it will consider they are in the parameter z.
    :param mode: Which antisymmetric function is applied to
    aggregate the statistics from the original features and from the
    knockoffs. By default, the difference is used.
    """
    if z_tilde is None:
        z, z_tilde = np.split(z, 2)

    if mode == "max":
        return np.maximum(z, z_tilde) * np.sign(z - z_tilde)
    elif mode == "difference":
        return np.abs(z) - np.abs(z_tilde)
    elif mode == "log":
        return np.log(z / z_tilde)
    else:
        raise ValueError(
            f"Argument 'mode' can be 'max', 'difference' or 'log'. Found {mode}"
        )


class AntisymmetricStatistics(KnockoffStatistics):
    """
    Flip-sign knockoff statistics computed in two steps,
    through and antisymmetric function.

    :param antisymmetry: Which antisymmetric function is applied to
    aggregate the statistics from the original features and from the
    knockoffs. By default, the difference is used.
    """

    def __init__(self, antisymmetry="difference"):
        super().__init__()
        self.antisymmetry = antisymmetry

    def evaluate(self, X, X_tilde, y):
        return antisymmetric_knockoff_statistics(
            self.evaluate_antisymmetric(X, X_tilde, y), mode=self.antisymmetry
        )

    def evaluate_antisymmetric(self, X, X_tilde, y):
        pass


class EstimatorStatistics(AntisymmetricStatistics):
    """
    Flip-sign knockoff statistics

    :param estimator: Estimator which is fitted against [X, X_tilde]
    in order to compute the statistics.
    It must provide an attribute "coef_" after it is fitted.
    By default, the LassoCV estimator from Scikit-Learn is used.
    :param antisymmetry: Which antisymmetric function is applied to
    aggregate the statistics from the original features and from the
    knockoffs. By default, the difference is used.
    :param absolute: Whether or not to keep the coefficients of
    the estimator in absolute value. Defaults to True.
    """

    def __init__(
        self, estimator=None, antisymmetry="difference", absolute: bool = True
    ):
        super().__init__(antisymmetry=antisymmetry)

        if estimator is None:
            estimator = LassoCV(cv=3)
        self.estimator = estimator
        self.absolute = absolute

    def evaluate_antisymmetric(self, X, X_tilde, y):
        stats = self.estimator.fit(np.hstack((X, X_tilde)), y).coef_
        if self.absolute:
            stats = np.abs(stats)
        return stats
